import { client, gateway, sessionManager, voiceStateManager } from "./api.js"

import {
	ActivityType,
	APIChatInputApplicationCommandInteraction,
	ApplicationCommandType,
	GatewayDispatchEvents,
	InteractionType,
	PresenceUpdateStatus,
	WithIntrinsicProps,
} from "@discordjs/core"
import { startCommand } from "./commands/start.js"
import { config } from "./config.js"
import { syncCommands } from "./utils/commands.js"
import { VoiceState } from "./utils/wrapper/VoiceState.js"

process.on("uncaughtException", (error) => console.error(error))
process.on("unhandledRejection", (error) => console.error(error))

if (config.sync) {
	await syncCommands()
}

client.on(GatewayDispatchEvents.GuildCreate, (args) => {
	console.log(`ADD GUILD: ${args.data.name}(${args.data.id})`)
	voiceStateManager.addGuild(args.data.id, args.data.voice_states)
})

client.on(GatewayDispatchEvents.GuildDelete, (args) => {
	voiceStateManager.removeGuild(args.data.id)
})

client.on(GatewayDispatchEvents.Ready, async (ready) => {
	console.log("ready", ready.data.user.username)

	await Promise.all(
		(
			await gateway.getShardIds()
		).map((shard) =>
			client
				.updatePresence(shard, {
					afk: false,
					since: Date.now(),
					status: PresenceUpdateStatus.Online,
					activities: [
						{
							type: ActivityType.Playing,
							name: "냥냥",
						},
					],
				})
				.then(() => {
					console.log("updated presence for shard", shard)
				})
		)
	)
})

client.on(GatewayDispatchEvents.InteractionCreate, async (args) => {
	if (args.data.type === InteractionType.ApplicationCommand) {
		if (args.data.data.type === ApplicationCommandType.ChatInput) {
			const command = args.data.data.name

			let commandFn:
				| ((
						args: WithIntrinsicProps<APIChatInputApplicationCommandInteraction>
				  ) => Promise<void>)
				| null = null

			switch (command) {
				case "start":
					commandFn = startCommand
					break
			}

			if (!commandFn) return

			await commandFn(
				args as WithIntrinsicProps<APIChatInputApplicationCommandInteraction>
			)
		}
	}
})

client.on(GatewayDispatchEvents.VoiceStateUpdate, async (args) => {
	if (args.data.guild_id) {
		const adapter = sessionManager.adapters.get(args.data.guild_id)
		if (adapter) adapter.onVoiceStateUpdate(args.data)
		const guild = voiceStateManager.guilds.get(args.data.guild_id)
		if (guild) {
			const state = guild.get(args.data.user_id)
			if (state) {
				state.update(args.data)
			} else {
				guild.set(
					args.data.user_id,
					new VoiceState(args.data, args.data.guild_id)
				)
			}
		}
	}
})

client.on(GatewayDispatchEvents.VoiceServerUpdate, (args) => {
	const adapter = sessionManager.adapters.get(args.data.guild_id)
	if (adapter) adapter.onVoiceServerUpdate(args.data)
})

await gateway.connect()
